/* ============================================================
 * File  : harvester.cpp
 * Author: Eric Giesselbach <ericgies@kabelfoon.nl>
 * Date  : 2004-12-18
 * Description : scans url for links (streams)
 *
 * Copyright 2003 by Eric Giesselbach

 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published bythe Free Software Foundation;
 * either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ============================================================ */

#include <iostream>

#include <q3network.h>
#include <q3textstream.h>
#include <qtimer.h>
#include <qstringlist.h>
#include <q3dict.h>
#include <qregexp.h>
#include <q3url.h>
#include <qmutex.h>
#include <qapplication.h>
#include <qdatastream.h>
#include <qdom.h>
//Added by qt3to4:
#include <QCustomEvent>
#include <QEvent>

#include <sys/stat.h>

#include "harvester.h"


using namespace std;

QMutex mutex;

typedef int* pint;

/*
  Users can share the download directory:
  - assign all users to a common group (e.g. users)
  - create, set group, setgid a central downloads directory:
      cd /var
      mkdir downloads
      chgrp users downloads
      chmod 2775 downloads
  - if the directory contains files created before setting gid:
      chmod -R 0775 downloads
      chgrp -R users downloads
  - symlink the new directory as "SUBPATH"/.streamtuned/downloads
      cd $HOME/.streamtuned
      rm -Rf downloads (don't try this at home)
      ln -s /var/downloads .

  Note:
    1. upon selection of a item with STREAM_DL handler, wget will download the
       uri to the download dir.
    2. the downloading file will be renamed as [filename]_temp
    3. when download finishes the file will be renamed as filename
    4. the process that started wget is aware of the running download,
    5. other streamtuned processes will regard the [filename]_temp as a
       failed download and will start a new download upon selection. What
       happens next is specified in the wget documentation (I guess).
*/


// only works if loaded in setup plugin
Downloader::Downloader(QString directory, StreamStorage *storage)
{
   this->directory = directory;
   this->home = getenv("HOME");
   
   downloads.clear();
   umask(2);  // read/write rights for user and group
   streamStorage = storage;
       connect(
             streamStorage,
             SIGNAL(recordRemoved(ChangedRecord*)),
             this,
             SLOT(slotRecordRemoved(ChangedRecord*)) );

}

QString Downloader::download(QString url, QString sName, bool& canplaynow, bool& error,
                             bool& itemcreated, QString& errorMessage, bool checkName)
// returns absolute path of downloading file
{
   QString path;
   Availability availability;
   bool pathFailed = false;

   error        = false;
   errorMessage = "";
   canplaynow   = false;
   itemcreated  = false;
   
   path = checkAvailable(url, availability, pathFailed, true, sName, checkName);

   // todo: when [filename]_temp file exists and no download on that file
   // is active, it should resume (wget -c). But Wget refuses to download if
   // server doesn't support resumes. Create workaround using wget feedback...
   //
   // For now: if "pending" check for running download and if none, ignore the _temp file.
   if ( availability == pending && downloads.find(url) == downloads.end() )
     availability = notavailable;
   
   canplaynow = ( availability != notavailable );
   if (pathFailed)
     errorMessage = " error: could not create file " + path;

   error = pathFailed;
   
   if (availability == notavailable && !pathFailed)
   {
        proc = new Q3Process(this);
        
        proc->addArgument( "wget" );
        proc->addArgument( "-O" );
        proc->addArgument( path );  // pos 2, required slotDownloadStopped
        proc->addArgument( url );   // pos 3, required slotDownloadStopped

        connect(
           proc,
           SIGNAL( processExited() ),
           this,
           SLOT( slotDownloadStopped() )
        );
        QString temp = ""; 
        error = !proc->launch(temp);

        if (error)
        {
          errorMessage = "Cannot create download process";
          delete proc;
        }
          else
        {
            downloads.append(url);
            itemcreated = createStreamItem(sName, path, url, "");  // filename passed as url
        }

   }

   return path;
}

//  download as path/filename_temp
//  rename to path/filename on succesful exit
//  resume download if path/filename_temp exists

void Downloader::slotDownloadStopped()
{
   bool error = true;
   QString errorMessage = "";
   
   QObject *obj = (QObject*)sender();

   if (obj)
   {
      Q3Process *m_proc = (Q3Process*)obj;
      
      QStringList args = m_proc->arguments();
      QString path = args[2];
      QString url  = args[3];
      
      if ( m_proc->normalExit() && m_proc->exitStatus() == 0 )  // download finished...
      {
         error = false;
         // rename filename_temp --> filename
         if ( path.right(5) == "_temp" )
         {
            QString newPath = path.left( path.length() - 5 );
            QDir dir(home);
            dir.rename(path, newPath);
            updateStreamItemFilename(path, newPath);
         }
      }  // else leave file for download resume/overwrite on next select
        else
          errorMessage = "Download failed (exit " + QString::number(m_proc->exitStatus()) + "): " + url;

      downloads.remove(url);
      delete m_proc;
      
      emit downloadFinished(url, error, errorMessage);
   }  
   
   //normalExit() // true if not aborted / crashed
   //exitStatus() // process return value
}

QString Downloader::checkAvailable( QString url, Availability& availability,
                                    bool& errorWhileCreating, bool createPathIfNotExists,
                                    QString name, bool checkName)
{
   QString path;
   errorWhileCreating = false;
   
   availability = notavailable;
   errorWhileCreating = false;
          
   QString urlPath = url.section( "://", 1, 1);  // strip protocol
 
   // change "../" to "xx/".  This prevents any unintended directory traversal.
   urlPath.replace("../", "xx/");
   
   path = directory + "/" + urlPath;
   QFile file(home + "/" + path);

   if ( file.exists() )
   {
     availability = available;
     return path;
   }

   path    += "_temp";
   urlPath += "_temp";
   file.setName(home + "/" + path);

   if ( file.exists() )
   {
     availability = pending;
     return path;
   }

 // should check host part of url (that should be, but is not?) in descr...
   if (checkName)
   {
		bool found = false;
		
		streamStorage->resetRecordList();

		ValueList values(5);
		while ( !found && streamStorage->getNextRecord(values) )
		{
			if ( values.count() == 5 && values[s_folder] == "downloads" && values[s_name] == name )
			found = true;
		}

		if (found)
		{
         availability = available;
         return values[s_url];
		}
   }

   if ( !createPathIfNotExists )
     return "";

   QStringList pathElements = QStringList::split( "/", urlPath );  // get directories and filename
   
   QDir dir;
   QString currentDir = directory;
   bool pathCreated = true;
   
   for ( int i = 0; i < pathElements.size() - 1; i++ ) // assume last entry is file
   {
      currentDir += "/" + *pathElements.at(i);
      //cout << "create: " << currentDir << endl;
      dir = QDir( home + "/" + currentDir );
      
      if ( !dir.exists() )
      {
        pathCreated = dir.mkdir(home + "/" + currentDir);
        if (!pathCreated) cout << "TARGET error: cannot create " << currentDir.latin1() << endl;
      }
   }

   errorWhileCreating = !pathCreated;
   
   if (errorWhileCreating)
     cout << "TARGET error: path create error for " << urlPath.latin1() << endl;
   
   return path;
}


bool Downloader::createStreamItem(QString name, QString url, QString descr, QString handler)
{
    QString errorMsg;
    ValueList values(5);
    values[s_folder] = "downloads";
    values[s_name]   = name;
    values[s_url]    = url;
    values[s_descr]  = descr;
    values[s_handler]  = handler;

    if ( !streamStorage->insertRecord(106, values, errorMsg) )
    {
       cout << "TARGET download create error: " << errorMsg.latin1() << endl;
       return false;
    }
       else return true;
}

bool Downloader::updateStreamItemFilename(QString oldFile, QString newFile)
{    
    QString errorMsg = "";
    
    bool found = false;
    
    streamStorage->resetRecordList();

    ValueList values(5);
    while ( !found && streamStorage->getNextRecord(values) )
    {
      if ( values.count() == 5 && values[s_folder] == "downloads" && values[s_url] == oldFile )
        found = true;
    }

    if (!found)
    {
      cout << "TARGET error: download stream item url " << oldFile.latin1() << " not found" << endl;
      return false;
    }
    
    ValueList ovalues = values;

    values[s_url] = newFile;
    
    if ( !streamStorage->updateRecord(106, ovalues, values, errorMsg) )
    {
       cout << "TARGET download update error: " << errorMsg.latin1() << endl;
       return false;
    }
       else return true;
}

bool Downloader::deleteStreamItem(QString name, QString url, QString descr, QString handler)
{
    QString errorMsg;
    ValueList values(5);
    values[s_folder] = "downloads";
    values[s_name]   = name;
    values[s_url]    = url;
    values[s_descr]  = descr;
    values[s_handler]  = handler;

    if ( !streamStorage->removeRecord(106, values, errorMsg) )
    {
       cout << "TARGET download delete error: " << errorMsg.latin1() << endl;
       return false;
    }
       else return true;
    
}

void Downloader::slotRecordRemoved(ChangedRecord* rec)
{
  if (rec->ident == 106) return;

  // delete file if: folder=downloads
  if ( rec->oldValues[0] == "downloads" )
  {
     //stopRecording(rec->values[1]);
     QFile(rec->oldValues[2]).remove();
  }
}



//----------------------------------------------------------------------------
// using cache is nice when working playlists, required when working rss/rdf feeds


CacheItem::CacheItem(QString lastdate, QString filename)
{
   this->filename = filename;
   this->lastdate = lastdate;
}

Cache::Cache()
{
  fileIndex = 10000;
  cachePath = QString(getenv("HOME")) + "/."SUBPATH"/cache";
  
  history.setAutoDelete(true);
  history.resize(1000);

  loadCache();
}

Cache::~Cache()
{
  saveCache();
}

bool Cache::copyCacheFile(QString source, QString dest)
{
  QFile src(source);
  QFile dst(dest);

  if ( !src.open(QIODevice::ReadOnly) )
  {
    cout << "TARGET error: cannot open " << source.latin1() << endl;
    return false;
  }
  if ( !dst.open(QIODevice::WriteOnly) )
  {
    cout << "TARGET error: cannot open " << dest.latin1() << endl;
    return false;
  }

  QTextStream srcStream( &src );
  QTextStream dstStream( &dst );
  QString line = "";

  while ( !srcStream.atEnd() )
  {
     line = srcStream.readLine(); // line of text excluding '\n'
     dstStream << line << endl;
  }
  
  dst.close();
  src.close();

  return true;
}


bool Cache::getCacheItem(QString url, QString& lastdate, QString& filename)
{
  lastdate = "";
  filename = "";
  
  CacheItem* item = history[url];
  
  if ( item )
  {
    lastdate = item->lastdate;
    filename = item->filename;
    return true;
  }
  else
    return false;
}


QString Cache::createNewFile()
{
    QString filename;
    QFile file;
    bool result = false;
    int attempt = 0;
    
    do
    {
      fileIndex++;
      filename = cachePath + "/cache" + QString::number(fileIndex);

      file.setName( filename );
      if ( !file.exists() )
      {
         result = file.open(QIODevice::WriteOnly);
         if ( file.isOpen() ) file.close();
         attempt++;
      }
    }
    while (!result && attempt < 5);

    if (result)
      return filename;
    else
      return "";
}


bool Cache::setCacheItem(QString url, QString lastdate, QString tempfilename)
{
  QString filename;
  CacheItem* item = history[url];
  
  if ( item )
  {
    item->lastdate = lastdate;
    copyCacheFile(tempfilename, item->filename);
  }
    else
  {
    filename = createNewFile();
    
    if (filename == "")
    {
       cerr << "cannot create cache file" << endl;
       return false;
    }
    
    copyCacheFile(tempfilename, filename);
    history.insert( url, new CacheItem(lastdate, filename) );
  }
  
  return true;
}

void Cache::loadCache()
{
    if ( !openCacheFile(false) )
      return;
    
    myFile.at(0);
    Q3TextStream stream( &myFile );

    history.clear();

    QString m_url = "";
    QString m_lastdate = "";
    QString m_filename = "";
    
    int stage = 0;
    QString line;

    while ( !stream.eof() )
    {
       line = stream.readLine();

       if (stage > 0) stage++;

       if ( line == "[item]" )
       {
         if (stage > 1) // data received
            history.insert( m_url, new CacheItem(m_lastdate, m_filename) );
         
         stage = 1;
       }

       // ignore empty lines, interpret [emptystring] as empty string.
       if ( line == "[emptystring]" ) line = "";
       
      
       if (line != "")
         switch (stage)
         {
           case 2: // url
             m_url = line;
           break;
           case 3: // lastdate
             m_lastdate = line;
           break;
           case 4: // filename
             m_filename = line;
           break;
           default:
           break;
         }
    }

    if ( stage > 1 )  // data received
      history.insert( m_url, new CacheItem(m_lastdate, m_filename) );

    closeCacheFile();
}

void Cache::saveCache()
{
    QString value;
    Q3TextStream stream( &myFile );
    CacheItem *item;
        
    openCacheFile(true);
    
    Q3DictIterator<CacheItem>i(history);

    for (; i.current(); ++i )
    {
      item = i.current();
      stream << endl;
      stream << "[item]" << endl;
      stream << i.currentKey() << endl;
      stream << item->lastdate << endl;
      stream << item->filename << endl;
    }

    closeCacheFile();
}

bool Cache::openCacheFile(bool overwrite)
{
   bool result = false;
   
   QString path = cachePath + "/cacheindex";
    
   myFile.setName( path );
   
   if (!overwrite)
     result = myFile.open(QIODevice::ReadOnly);
   
   if (!result)
     result = myFile.open(QIODevice::WriteOnly);

   if (!result)
     cerr << "cannot open file " << path.latin1() << endl;
      
   return result;
}

bool Cache::closeCacheFile()
{
   if ( myFile.isOpen() )
   {
     myFile.close();
     return true;
   }
     else
       return false;
}


//----------------------------------------------------------------------------

FetchBuffer::FetchBuffer(QString host, QString url, QString fileName, int requestId, QString& errorMessage)
{
    errorMessage = "";
    responseDate = "";
    urlEstimateCount = 0;
    
    this->requestId = requestId;
    this->url = url;
    this->host = host;
    this->fileName = fileName;
    
    file.setName(fileName);
    
    if ( !file.open(QIODevice::WriteOnly) )
      errorMessage = "cannot open file " + fileName;
    else
    {
      stream.setDevice( &file );
      stream.setEncoding(QTextStream::UnicodeUTF8);
    }
}


FetchBuffer::~FetchBuffer()
{
   closeBuffer();
}


void FetchBuffer::closeBuffer()
{
  
   if ( file.isOpen() )
     file.close();
}
     
void FetchBuffer::writeData(QString& text)
{
   text = QString::fromUtf8(text);

   if ( file.isOpen() )
     stream << text;
}


//----------------------------------------------------------------------------

// "set-cookie"-line from response header
void CookieBin::updateCookie(QString host, QString newCookie)
{
    //cout << "process cookie: " << newCookie << " for host: " << host << endl;
    
    // For now: ignore domain (use host instead), expiration, secure flag
    QStringList newCookieParts = QStringList::split( ";", newCookie );
    newCookie = newCookieParts[0];

    //cout << "  new cookie: " << newCookie << endl;
    
    QString name  = newCookie.section( '=', 0, 0 ).stripWhiteSpace();
    QString value = newCookie.section( '=', 1, 1 ).stripWhiteSpace();

    newCookie = name + "=" + value;
    
    // first cookie for host
    if ( !hasCookies(host) )
    {  
      cookieMap.insert(host, newCookie);
      return;
    }
      
    QStringList cookieParts = QStringList::split( "; ", cookieMap[host] );

    int i = 0;
    bool processed = false;
    
    while ( i < cookieParts.size() && !processed )
    {
      // cookie update
      if ( cookieParts[i].startsWith(name + "=") )
      {
        processed = true;
        
        if (value == "")
          cookieParts.remove( cookieParts[i] );
        else
          cookieParts[i] = newCookie;
      }
      
      i++;
    }

    if (!processed && value != "")
      cookieParts.append(newCookie);
      
    cookieMap[host] = cookieParts.join("; ");
    //cout << "COOKIES STORED: " << cookieMap[host] << endl;
}

// have cookie(s) for host?
bool CookieBin::hasCookies(QString host)
{
   return cookieMap.contains(host);
}  

// "cookie"-line to send with request header
QString CookieBin::getCookieHdrLine(QString host)
{
   if ( hasCookies(host) )
     return cookieMap[host];
   else
     return QString::null;
}

void CookieBin::setReferer(QString host, QString referer)
{
   this->referer = referer;
   refererHost = host;
   //cout << "setting referer: " << referer << " for host " << host << endl;
}

QString CookieBin::getReferer(QString host)
{
   //cout << "get referer for host: " << host << " while refererHost = " << refererHost << endl;
   if ( refererHost != host )
     return "";
   else
     return referer;
}

bool CookieBin::hasReferer(QString host)
{
   return (refererHost == host);
}

//----------------------------------------------------------------------------


Requester::Requester() : QHttp()
{
  http = 0;
  cache = new Cache();

  http = new QHttp();
  fetchBuffer = 0;
  
  connect(
           http,
           SIGNAL(requestFinished(int, bool)),
           this,
           SLOT(slotRequestFinished(int, bool))
         );

  connect(
           http,
           SIGNAL(requestStarted(int)),
           this,
           SLOT(slotRequestStarted(int))
         );

   connect(
           http,
           SIGNAL(dataReadProgress(int, int)),
           this,
           SLOT(slotDataReadProgress(int, int))
         );

  connect(
           http,
           SIGNAL(readyRead(const QHttpResponseHeader&)),
           this,
           SLOT(slotReadyRead(const QHttpResponseHeader&))
         );

  connect(
           http,
           SIGNAL( responseHeaderReceived(const QHttpResponseHeader&) ),
           this,
           SLOT( slotResponseHeaderReceived(const QHttpResponseHeader&) )
         );
}


Requester::~Requester()
{
  if (http) delete http;
  if (cache) delete cache;
  if (fetchBuffer) delete fetchBuffer;
}


bool Requester::fetchData(QString url, QString fileName, QString& errorMsg)
{
//
    port = 0;
    host = "";
    this->url = "";
    this->fileName = "";
    redirectCount = 0;
    hHost = "";
    hUrl = "";
//
    status = abort;
    http->abort(); // Synchr. call to slotRequestFinished w. error set
    
    redirectCount = 0;
    this->fileName = fileName;
        
    Q3Url *qurl = new Q3Url( url );

    if (!qurl->isValid() || !qurl->hasHost() || qurl->protocol() != "http")
    {
      delete qurl;
      return false;
    }


    if (fetchBuffer)
    {
      delete fetchBuffer;
      fetchBuffer = 0;
    }
    
    hHost = qurl->host();
    port  = qurl->port();

    if (port != -1)
      hHost += ":" + QString::number(qurl->port());
    else
      port=80;

    hUrl = qurl->encodedPathAndQuery();
    host = qurl->host();
    delete(qurl);

    this->url = url;

    status = sethost;
    
    int ident = http->setHost(host, port);
    //cout << "HTTP host: " << qPrintable(host) << " got id = " << ident << endl;
    
    fetchBuffer = new FetchBuffer( host, url, fileName, ident, errorMsg);
        
    if (errorMsg != "")
    {
       delete fetchBuffer;
       fetchBuffer = 0;
       return false;
    }

    return true;
}


void Requester::slotResponseHeaderReceived(const QHttpResponseHeader& response)
{
    QString cachefilename, cachelastdate, temp;
    
   // check for cookies
   // bow for QT4: FINALLY response.values is a list instead of a map so we can get multiple cookies from the header
    QList< QPair<QString,QString> > list = response.values();
    QPair<QString,QString> pair;
    
    for (int i = 0; i < list.count(); i++)
    {
        pair = list.at(i);
        if (pair.first.compare("set-cookie", Qt::CaseInsensitive) == 0)
        {
         //cout << "FEED COOKIE: " << qPrintable(pair.second) << endl;
         cookieBin.updateCookie( host, pair.second );
        }
    }
    
    if ( response.statusCode() == 304 ) // Not Modified
    {
      //cout << "response: not modified" << endl;
      fetchBuffer->setResponseDate( "CACHE" );
      status = notmodified;
      http->abort();
    }
       // patch from Andreas Gerstlauer, handles redirects
       else if ( response.statusCode() >= 300 && response.statusCode() <= 400 ) // redirect
    {
      status = sethost;

      this->url = response.value("Location");
      //cout << "redirected to " << qPrintable(url) <<  endl;
           
      Q3Url *qurl = new Q3Url( url );

      status = abort;
      http->abort(); // Synchr. call to slotRequestFinished w. error set
      //TODO: reinit of http should not be required
      // without reinit, connection sends RST packets directly after GET <redirect URL>
  http = new QHttp();
  fetchBuffer = 0;
  
  connect(
           http,
           SIGNAL(requestFinished(int, bool)),
           this,
           SLOT(slotRequestFinished(int, bool))
         );

  connect(
           http,
           SIGNAL(requestStarted(int)),
           this,
           SLOT(slotRequestStarted(int))
         );

   connect(
           http,
           SIGNAL(dataReadProgress(int, int)),
           this,
           SLOT(slotDataReadProgress(int, int))
         );

  connect(
           http,
           SIGNAL(readyRead(const QHttpResponseHeader&)),
           this,
           SLOT(slotReadyRead(const QHttpResponseHeader&))
         );

  connect(
           http,
           SIGNAL( responseHeaderReceived(const QHttpResponseHeader&) ),
           this,
           SLOT( slotResponseHeaderReceived(const QHttpResponseHeader&) )
         );

      if (fetchBuffer)
      {
        delete fetchBuffer;
        fetchBuffer = 0;
      }

      if (!qurl->isValid() || redirectCount > 20) // no hashost, allow relative path
      {  
        delete qurl;

        if (redirectCount > 20)
          temp = TARGET": too much redirections, giving up";
        else
          temp = TARGET": redirected to invalid URL";
        //cout << temp << endl;
        emit fetchReady(true, temp);

        return;
      }
     
      if (qurl->hasHost())
      {
         hHost = qurl->host();
         port  = qurl->port();
         host  = qurl->host();

         if (port != -1)
           hHost += ":" + QString::number(qurl->port());
         else
           port=80;
      }
    
      hUrl = qurl->encodedPathAndQuery();
      delete(qurl);
      
      status = sethost;
      redirectCount++;
      QString errorMsg;
      
      int ident   = http->setHost(host, port);         
      //cout << "HTTP host(redirect): " << qPrintable(host) << " got id = " << ident << endl;
      fetchBuffer = new FetchBuffer( host, url, fileName, ident, errorMsg);
      
      //cout << "redirecting" << endl;
      
      if (errorMsg != "")
      {
         status = abort;
         http->abort();
         delete fetchBuffer;
         fetchBuffer = 0;
         cout << "TARGET error: " << errorMsg.latin1() << endl;
         return;
      }

    }
    
    if ( response.statusCode() == 200 && fetchBuffer ) // get reported date
      fetchBuffer->setResponseDate( response.value("Date") );
}

void Requester::slotDataReadProgress(int read, int size)
{
   if ( !fetchBuffer )
   {
      cerr << "no buffer!" << endl;
      return;
   }

   if ( read > 25000 && (fetchBuffer->urlEstimateCount == 0 ||
                         int(read/fetchBuffer->urlEstimateCount) > 50000 ) )
   {
     cout << "ABORT fetch: no url's detected" << endl;
     http->abort();
     return;
   }

   emit fetchProgress(read, size);
}

void Requester::slotReadyRead(const QHttpResponseHeader&)
{
    QString temp = QString( http->readAll() );
    
    if ( !fetchBuffer )
    {
       cerr << "no buffer to write to " << endl;
       return;
    }

    if ( fetchBuffer->isOpen() )
    {
      fetchBuffer->writeData(temp);
    }
    else
      cerr << "stream feed buffer not open " << endl;
    
   // --> prevent read of massiva amounts of binary data
   // count the number of complete url's in this chunk
   // add to total, do some math and decide to continue

   // can be wrong, but code below is useless since default parser delegation?
   //cout << "slotReadyRead" << endl;
    // check for protocol://server.domain
    int index = 0;
    QRegExp expr;
    expr.setPattern("\\w+://[\\w\\-]+\\.[\\w\\-]+"); // matchprobability in noise: a few 1E-6
    
    while ( index > -1)
    {
      index  = expr.search( temp, index );
      index += expr.matchedLength();
      if ( index > -1 ) fetchBuffer->urlEstimateCount++;
    }

    expr.setPattern("<a\\s+href"); // matchprobability in noise: dunno
    index = 0;
    
    while ( index > -1)
    {
      index  = expr.search( temp, index );
      index += expr.matchedLength();
      if ( index > -1 ) fetchBuffer->urlEstimateCount++;
    }
    
}

void Requester::slotRequestStarted(int /*id*/)
{
    //cout << "HTTP Request " << id << " started" << endl;
}

void Requester::slotRequestFinished(int id, bool error)
{
    //cout << "HTTP Request " << id << " finished" << endl;
    QString errorMsg;
    QString cachefilename, cachelastdate, temp;
    
    if ( status == abort || status == notmodified ) // our abort
      error = false;
    
    if ( http->error() != QHttp::NoError || error ) // http error
      errorMsg = "HTTP error: " + http->errorString();
  
    switch ( status )
    {
      case sethost:  // host lookup finished
            //cout << "... was setHost request" << endl;
            if (!error && fetchBuffer)
            {
              status = getdata;
              
              cache->getCacheItem(url, cachelastdate, cachefilename);
              QHttpRequestHeader header( "GET", hUrl );
              
              if ( cookieBin.hasCookies(host) )
                header.setValue( "Cookie", cookieBin.getCookieHdrLine(host));
                
              if ( cookieBin.hasReferer(host) )
                header.setValue( "Referer", cookieBin.getReferer(host) );

              cookieBin.setReferer(host, hHost + hUrl);

              if (cachelastdate != "")
                header.setValue( "If-Modified-Since", cachelastdate );

              //Accept: text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5

              header.setValue( "User-Agent", "Mozilla/5.0 (compatible; MythStream-v"VERSION")" );
              header.setValue( "Connection", "close" );
              header.setValue( "Host", hHost );
              
              //cout << "... http request data: " << qPrintable(header.toString()) << endl;
              
              fetchBuffer->setRequestId( http->request( header ) );
            }
      break;
  
      case getdata:
            //cout << "... was getdata request" << endl;
            if ( !error )
            {
              if ( fetchBuffer && fetchBuffer->getRequestId()==id && fetchBuffer->isOpen() )
              {
                if ( http->bytesAvailable() )
                {
                  temp = QString(http->readAll());
                  fetchBuffer->writeData(temp);
                }
                
                fetchBuffer->closeBuffer();

                if ( fetchBuffer->getResponseDate() == "CACHE" )
                {
                    cache->getCacheItem(url, cachelastdate, cachefilename);
                    cache->copyCacheFile( cachefilename, fetchBuffer->getFileName() );
                }
                    else
                      cache->setCacheItem( fetchBuffer->getUrl(), fetchBuffer->getResponseDate(),
                                           fetchBuffer->getFileName() );
              
                delete fetchBuffer;
                fetchBuffer = 0;
              }
                else
                    cerr << "no stream feed buffer" << endl;
                
                            
              emit fetchReady(false, errorMsg);
            }
              //else 
                //cout << "... showing an error: " << qPrintable(http->errorString()) << endl;
      break;
  
      case notmodified:
            //cout << "... request received notmodified" << endl;
            if ( !error )
            {
            if ( fetchBuffer && fetchBuffer->getRequestId()==id && fetchBuffer->isOpen() )
              {
                if ( fetchBuffer->getResponseDate() == "CACHE" )
                {
                    cache->getCacheItem(url, cachelastdate, cachefilename);
                    cache->copyCacheFile( cachefilename, fetchBuffer->getFileName() );
                }

                delete fetchBuffer;
                fetchBuffer = 0;
              }
                else
                    cerr << "no stream feed buffer" << endl;
                            
              emit fetchReady(false, errorMsg);
            }

      break;
      
      case abort:
              //cout << "... request was aborted" << endl;
              error = true;
              errorMsg = "fetch aborted";
      break;
  
      default:
            //cout << "... request was WHAT??" << endl;
      break;
    }

    if (error)
    {
      if ( fetchBuffer && fetchBuffer->getRequestId() == id )
      {
        delete fetchBuffer;
        fetchBuffer = 0;
      }
      emit fetchReady(true, errorMsg);
    }

}

void Requester::stop()
{
   http->abort();
}

//----------------------------------------------------------------------------

NewUrl::NewUrl(QString nUrl, QString nName)
{
    url   = nUrl;
    name  = nName;
    proto   = "";
    port    = -1;
    file    = "";
    context = "";
    tag     = 0;
}

StreamHarvester::StreamHarvester()
{
    currentUrl   = "";
    currentName  = "";
    currentDescr = "";
    currentHandler = "";

    defaultParser = "";
    
    currentTag   = 0;

    history.setAutoDelete(true);
    history.clear();

    urlList.setAutoDelete(true);
    urlList.clear();
    rootTag = 0;

    eParser = NULL;
    
    busy    = false;
    pending = false;
    aborted = false;

    QString home = getenv("HOME");
    parsersPath = home + "/."SUBPATH"/parsers/";
    parsersPath_install = QString(PREFIX"/share/"SUBPATH"/parsers/");
    
    //cout << "parsers path: " << parsersPathi.latin1() << endl;
    
    requester = new Requester();
    connect ( requester, SIGNAL(fetchReady(bool, QString&)), this, SLOT(slotfetchReady(bool, QString&)) );
    connect ( requester, SIGNAL(fetchProgress(int, int)), this, SLOT(slotFetchProgress(int, int) ) );
}

StreamHarvester::~StreamHarvester()
{
   delete requester;
   if (eParser)
   {
      externalParserTerminate();
      delete(eParser);
   }
}


// ****************************** public ******************************

void StreamHarvester::fetchData(QString url, QString name, QString descr, QString handler)
{
    //cout << "fetch call" << endl;
    aborted = true;
    pending = true;

    currentName  = name;
    currentUrl   = url;
    currentDescr = descr;
    currentHandler = handler;
    
    currentTag   = 0;

    if (busy)
      stopFetch();
    else
      startFetch();
}

void StreamHarvester::parseData(QString& data, QString& name, QString& descr, QString& handler)
{
    if ( busy ) return;

    aborted = false;
    pending = false;
    busy    = true;

    currentName  = name;
    currentUrl   = "";
    currentDescr = descr;
    currentHandler = handler;
    
    currentTag   = 0;

    // skip data retrieval process
    this->data = data;
    processExited();
}

void StreamHarvester::stop()
{
   requester->stop();
   externalParserTerminate();
}

bool StreamHarvester::isBusy()
{
    return busy;
}

QString& StreamHarvester::getCurrentUrl()
{
   return currentUrl;
}

QString& StreamHarvester::getCurrentName()
{
   return currentName;
}

QString& StreamHarvester::getCurrentDescr()
{
   return currentDescr;
}

QString& StreamHarvester::getCurrentHandler()
{
   return currentHandler;
}

int StreamHarvester::getCurrentTag()
{
   return currentTag;
}

bool StreamHarvester::goBack()
{
   mutex.lock();

   history.removeLast();
   NewUrl *url = history.last();

    if ( url )
    {
      currentUrl   = url->url;
      currentName  = url->name;
      currentDescr = url->values["descr"];
      currentHandler = url->values["handler"];
      
      history.removeLast();

      mutex.unlock();

      fetchData(currentUrl, currentName, currentDescr, currentHandler);
      currentTag = url->tag;
      return true;
    }
      else
    {
      mutex.unlock();
      return false;
    }
}

void StreamHarvester::setTagOnCurrent(int tag)
{
   mutex.lock();

   NewUrl *url = history.last();
   if ( url )
     url->tag = tag;
   else
     rootTag = tag;

   mutex.unlock();
}

// ****************************** private ******************************

void StreamHarvester::startFetch()
{
    data = "";
    QString errorMessage = "";
    //slotfetchReady(bool, QString&)
    if ( currentHandler[0] == '*') // standalone parser, do not fetch with requester
    {
      //cout << "standalone parser " << currentHandler << endl;
      pending = false;
      aborted = false;
      processExited();
      return;
    }
    
    // currentHandler
    if ( requester->fetchData(currentUrl, parsersPath + "list.xml", errorMessage) )
    {
      busy = true;
      emit fetchStatus(receiving, 0);
    }
      else
        emit fetchStatus(errorset, invalidurl);
}


void StreamHarvester::stopFetch()
{
   //cout << "stop fetch" << endl;
   externalParserTerminate();
   startFetch();
}


// ************ external process callback **************


ParserType StreamHarvester::checkParserType(QString &parser, QString &path)
{
    //PERL
    QFileInfo fileInfo;

    fileInfo = QFileInfo( path + parser + ".pl" );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".pl";
        return PERL;
    }

   //SHELL
    fileInfo = QFileInfo( path + parser + ".sh");
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".sh";
        return SHELL;
    }

    //EXEC
    fileInfo = QFileInfo( path + parser );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
        return EXEC;
            
    //PYTHON
    fileInfo = QFileInfo( path + parser + ".py" );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".py";
        return PYTHON;
    }

    return NONE;
}

ParserType StreamHarvester::getParser(QString &parser, QString &path)
{
    QString nParser = parser;
    ParserType parserType;
    
   if ( nParser[0] == '*') nParser.remove(0,1); // remove standalone parser token *
   
   parserType = checkParserType(nParser, path);
    
    if ( parserType != NONE )
    {
        parser = nParser;
        return parserType;
    }

    parserType = checkParserType(nParser, parsersPath_install);

    if ( parserType != NONE )
    {
        parser = nParser;
        path = parsersPath_install;
        return parserType;
    }
    

    cerr << "external parser not found: " << nParser.latin1() << endl;

    // fallback to default
    if ( nParser != defaultParser )
    {
      cerr << "falling back to " << defaultParser.latin1() << endl;
      parser = defaultParser;
      return getParser( parser, path );
    }

    return NONE;
}


void StreamHarvester::processExited()
{

   // user posted new fetch request
   if (pending)
   {
     startFetch();
     return;
   }

   // no new fetch, but stopFetch still called
   if (aborted)
   {
     aborted = false;
     busy    = false;
     return;
   }
   
   externalParserStart(currentUrl, currentHandler, currentName);
}

void StreamHarvester::slotFetchProgress(int read, int) // data read by Requester
{
   read = read / 1024; // assuming read in bytes
   emit fetchStatus(receiving, read);
}

void StreamHarvester::slotfetchReady(bool error, QString& data)
{
   this->data = data;
   pending = false;
   aborted = false;

   lastHost = requester->hHost;
   
   if (!error)
     processExited();
   else
   {
     busy = false;
     // leaving data (error message) unused
     if (data == "fetch stopped") // if due to pending fetch the resulting message will be overwritten
       emit fetchStatus(errorset, nourlindata);
     else
       emit fetchStatus(errorset, nodata);
   }
}

void StreamHarvester::customEvent(QCustomEvent *ev)
{
	//cout << "received event" << endl;
    if ( ev->type() == QEvent::User + StreamHarvester::ready )
    {
	//cout << "received event ready" << endl;
       int *cnt = (int*)ev->data();
       parserExited(*cnt);
       delete cnt;
    }

    if ( ev->type() == QEvent::User + StreamHarvester::parsing )
    {
	//cout << "received event parsing" << endl;
       int *cnt = (int*)ev->data();
       emit fetchStatus(parsing, *cnt);
       delete cnt;
    }
}

void StreamHarvester::parserExited(int matchCount)
{
   // parser ready (and not aborted)
   // not threadsafe
   
   mutex.lock();

   NewUrl *newUrl;
   
   if ( matchCount > 1 )
   {
      newUrl = new NewUrl( currentUrl, currentName);
      newUrl->values["descr"] = currentDescr;
      newUrl->values["handler"] = currentHandler;

      history.append(newUrl);

     if (history.count() > 50)
       history.removeFirst();
   }

   busy = false;

   mutex.unlock();
//cout << "sending fetchStatus ready" << endl;
   emit fetchStatus(ready, matchCount);
}

// outsourcing the parser factory...

void StreamHarvester::externalParserStart(QString srcUrl, QString parser, QString srcName)
{
    if ( eParser ) return;

    QString ppath = parsersPath;
    //cout << "default parser: " << defaultParser << " advised parser: " << parser << ext << endl;
    
    // empty parser field is implicit configuration of default parser
    if (parser == "") parser = "default";

    ParserType parserType = getParser(parser, ppath);
    
    sourceUrl = srcUrl;
    
    urlCount = 0;
    eParserUrl = NULL;
    urlList.clear();
    parserOutput = "";
    
    eParser = new Q3Process( this );  
    eParser->setWorkingDirectory(QDir(ppath)); // was bug
    eParser->setCommunication( Q3Process::Stdin |
                            Q3Process::Stdout
    //                        |QProcess::Stderr
    //                        |QProcess::DupStderr
                            );
    // The source Url must be quoted to avoid the shell
    // interpreting special url characters like &,, *, ...
    // The source Url can therefore not contain quotes itself.
    if ( sourceUrl.find('"') != -1 || sourceUrl.find('\\') != -1)
    {
      cerr << "Mythstream: insecure sourceURL: " << sourceUrl.latin1() << endl;
      sourceUrl = "";
    }
                  
    srcName = srcName.replace('"', "'");

    while (srcName.right(1) == "\\")
      srcName.truncate( srcName.length() );

// 	cout << "parser type: " << parserType << endl;
// 	cout << "parser: " << ppath.latin1()  << parser.latin1() << endl;
	
	switch (parserType)
	{
		case PERL:
    // execute perl script
    eParser->addArgument( "perl" );
    eParser->addArgument( ppath + parser );
    eParser->addArgument( "list.xml" );
    eParser->addArgument( "\"" + sourceUrl + "\"" );  // change
    eParser->addArgument( "\"" + srcName + "\"" );  // change
    //cout << "PERL PARSER" << endl;
	break;
	case EXEC:
    // executable binary
    eParser->addArgument( ppath + parser );
    eParser->addArgument( "list.xml" );
    eParser->addArgument( "\"" + sourceUrl + "\"");  // change
    eParser->addArgument( "\"" + srcName + "\"");  // change
    //cout << "EXEC PARSER" << endl;
	break;
	case SHELL:

    // execute any executable
    //
    // requires parser files to be executable!
    // parser and sourceUrl are external
    // trust parser: assume the perl script defines that value
    
    eParser->addArgument( "sh" );
    eParser->addArgument( "-c" );
    eParser->addArgument( ppath + parser + " list.xml \"" + sourceUrl + "\" \"" + srcName + "\"" );
    break;
		case PYTHON:
    // execute perl script
    eParser->addArgument( "python" );
    eParser->addArgument( ppath + parser );
    eParser->addArgument( "list.xml" );
    eParser->addArgument( "\"" + sourceUrl + "\"" );  // change
    eParser->addArgument( "\"" + srcName + "\"" );  // change
	break;
        default:
        break;
    }
    

    pending = false;

    connect( eParser, SIGNAL(readyReadStdout()),
            this, SLOT(externalParserRead()) );
    connect( eParser, SIGNAL(processExited()),
            this, SLOT(externalParserExited()) );

    
    if ( !eParser->start() )
      fprintf( stderr, "error starting parser " + parser + "\n" );
    
    baseUrl = sourceUrl;
    int lastSlash = baseUrl.findRev("/", -1);
    int lastDot   = baseUrl.findRev(".", -1);
    if ( 
         lastSlash > 0 && 
         lastSlash < lastDot &&
         baseUrl.mid(lastSlash - 1, 1) != "/"
       )
         baseUrl = baseUrl.left(lastSlash);
}

void StreamHarvester::externalParserTerminate()
{
   if ( eParser && eParser->isRunning() )
   {
      eParser->tryTerminate();
   }
}


void StreamHarvester::send(QString type, int cnt)
{
   int eventType = 0;
   if (type == "result"   ) eventType = QEvent::User + StreamHarvester::ready;
   if (type == "linkcount") eventType = QEvent::User + StreamHarvester::parsing;

   QCustomEvent *ev = new QCustomEvent(eventType);
   int *i = new int(cnt);
   ev->setData(i);
   //qApp->postEvent(this, ev);
   customEvent(ev);
}

void StreamHarvester::storeParserUrl()
{
        if ( eParserUrl ) // && eParserUrl->url != sourceUrl )
        {
          if (eParserUrl->name == "") eParserUrl->name = eParserUrl->url;
          urlList.append( eParserUrl );
          urlCount++;
          eParserUrl = NULL;
        }
          else 
        {
          delete(eParserUrl);
          eParserUrl = NULL;
        }
}


bool StreamHarvester::externalParserLoadUrl(QString& xmlData)
{ 
    int     port;
    QString protocol, file, context;
    Q3Url    *qurl;
    
    QDomDocument doc;    

    //cout << "BEGIN parser result:" << endl;
    //cout << xmlData << endl;
    //cout << "END parser result" << endl;
    
    if (!doc.setContent(xmlData, false) ) 
    {
       cerr << TARGET": could not read parser output: " << endl;
       cerr << "-------------------------------------" << endl;
       cerr << xmlData.latin1() << endl;
       cerr << "-------------------------------------" << endl;
       
       return false;
    }
    
    QDomElement docElem = doc.documentElement();
    QDomNode item = docElem.firstChild();
    QDomNode nName, nUrl, nDescr, nHandl, nValue, nMeta, nError1, nError2;
    QDomNodeList lMeta;
    QString name, url, descr, handl, value, error1, error2, type;
    QDomNode fName, fType, fValue;
    QDomElement eMeta;
    QString metaName, metaContent, metaViewer;
    
    while( !item.isNull() ) 
    {
        QDomElement e = item.toElement();
        
        if ( e.isNull() || (e.tagName() != "item" && e.tagName() != "error"
                            && e.tagName() != "field" ) )
        {
          cerr << TARGET": invalid parser output" << endl;
          return false;
        }

        if (e.tagName() == "error")
        {
            error1 = ""; error2 = "";
            nError1  = e.elementsByTagName("line1" ).item(0);
            nError2  = e.elementsByTagName("line2"  ).item(0);
            if ( !nError1.isNull() )
              error1 = nError1.toElement().text();
            if ( !nError2.isNull() )
              error2 = nError2.toElement().text();
              
            emit showError(error1, error2);
        }
            else if (e.tagName() == "field")
        {
            //cout << "field " << endl;
            fName  = e.elementsByTagName("name" ).item(0);
            fType  = e.elementsByTagName("type"  ).item(0);
            fValue = e.elementsByTagName("value").item(0);
            
            if ( !fName.isNull() && !fType.isNull() )  // required nodes
            {
                  name = fName.toElement().text();
                  type  = fType.toElement().text();
                  
                  if ( fValue.isNull() )  // value
                     value = "";
                  else
                     value = fValue.toElement().text();
                     
                  value = QString::fromUtf8(value);
                  emit playerServiceParameter(name, type, value);
            }
        }
            else  // e.tagName is "item"
        {
            nName  = e.elementsByTagName("name" ).item(0);
            nUrl   = e.elementsByTagName("url"  ).item(0);
            nDescr = e.elementsByTagName("descr").item(0);
            nHandl = e.elementsByTagName("handler").item(0);
            lMeta  = e.elementsByTagName("meta");
            
            if ( !nName.isNull() && !nUrl.isNull() )  // required nodes
            {           
                  name = nName.toElement().text();
                  url  = nUrl.toElement().text();
                  
                  if ( nDescr.isNull() )  // description (single line)
                     descr = "";
                  else
                     descr = nDescr.toElement().text();
                  
                  if ( nHandl.isNull() )  // handler (how to handle this item)
                     handl = "";
                  else
                     handl = nHandl.toElement().text();

                  eParserUrl = new NewUrl(url, name);
                  eParserUrl->values["descr"] = descr;
                  eParserUrl->values["handler"] = handl;
                                    
                  // player service
                  nValue  = e.elementsByTagName("pserv").item(0);
                  if ( nValue.isNull() ) value = ""; else value = nValue.toElement().text();
                  eParserUrl->values["pserv"] = value;
                  
                  //SUB subtitles
                  nValue  = e.elementsByTagName("subtitles").item(0);
                  if ( nValue.isNull() ) value = ""; else value = nValue.toElement().text();
                  eParserUrl->values["subtitles"] = value;

                  // logo url, displayed after item start
                  nValue  = e.elementsByTagName("logo").item(0);
                  if ( nValue.isNull() ) value = ""; else value = nValue.toElement().text();
                  eParserUrl->values["logo"] = value;
                  
                  context = "absolute";
                  // check if relative
                  if ( url.contains("://") == 0 )  // relative url
                  {
                     if ( url.find('/') == 0 )  //absolute path
                     {
                        url = "http://" + lastHost + url;
                     }
                        else
                     {
                        context = "relative";
                        url = baseUrl + "/" + url;
                        url = url.replace( QRegExp("[^/]+/\\.\\./"), "" ); // test ../../ recursive?
                        url = url.replace( QRegExp("\\./"), "" );
                        url = url.replace( "///", "/" );
                     }
                     
                  }
            
                  // get url properties
                  qurl = new Q3Url( url );
                  protocol = qurl->protocol();
                  port = qurl->port();
                  file = qurl->fileName() + qurl->query();
            
                  // block
                  // quick hack, block should be transported to streambrowser.cpp ...
                  if ( 
                        file.contains(".rm")  > 0 ||
                        file.contains(".ra")  > 0 ||
                        file.contains(".wav") > 0 ||
                        file.contains(".ram") > 0 ||
                        file.contains(".asx") > 0 ||
                        file.contains(".avi") > 0 ||
                        file.contains(".mpg") > 0 ||
                        file.contains(".pls") > 0 ||
                        file.contains(".mp3") > 0 ||
                        file.contains(".ogg") > 0 ||
                        file.contains(".wmf") > 0 ||
                        file.contains(".smi") > 0 ||
                        file.contains(".m3u") > 0 
                     )
                     file = "stream";
                  else
                     file = "other";
            
                  delete ( qurl );
                     
                  eParserUrl->proto   = protocol;
                  eParserUrl->file    = file;
                  eParserUrl->context = context;
                  eParserUrl->port    = port;
                  eParserUrl->url     = url;
                  //cout << "found name: " << name << ", handler: " << handl << endl;
                  // process meta information
                  for (int i = 0; i<lMeta.count(); i++)
                  {
                     nMeta = lMeta.item(i);
                     metaName = "";
                     
                     if ( !nMeta.isNull() )
                     {
                        eMeta = nMeta.toElement(); 
                        if ( !eMeta.isNull() )
                        {
                           metaName    = eMeta.elementsByTagName("name"    ).item(0).toElement().text(); // so what :)
                           metaContent = eMeta.elementsByTagName("content" ).item(0).toElement().text();
                           metaViewer  = eMeta.elementsByTagName("viewer"  ).item(0).toElement().text();
                        }
                     } 
                     
                     if ( metaName != "" ) 
                        eParserUrl->meta.append(metaName + "(stsep)" + metaViewer + "(stsep)" + metaContent);
                  }
                  
                  storeParserUrl();
            }

        }
                  
        item = item.nextSibling();
    }

    return true;
}        

void StreamHarvester::externalParserRead()
{
   QString temp = QString::null;
   
   do 
   {
     if (temp != QString::null) 
        parserOutput += QString::fromUtf8(temp);
     temp = eParser->readLineStdout();
     //cout << "read: " << qPrintable(temp) << endl;
   }
     while (temp != QString::null);
}


void StreamHarvester::externalParserExited()
{
   bool success = externalParserLoadUrl(parserOutput);
   if (eParser) delete eParser;
   eParserUrl = NULL;
   eParser = NULL;
   //cout << "tot urls: " << urlCount << endl;
   if (success)
   {
     //cout << "parser ok" << endl;
     send("result", urlCount);
   }
   else
     emit fetchStatus(errorset, bogusparseroutput);
}

//=============================================================================
//=============================================================================
//=============================================================================


PlayerService::PlayerService()
{
    eParser = NULL;
    error = "";

    parsersPath = QString(getenv("HOME")) + "/."SUBPATH"/parsers/";
    parsersPath_install = QString(PREFIX"/share/"SUBPATH"/parsers/");
}

PlayerService::~PlayerService()
{
   if (eParser)
   {
      externalParserTerminate();
      delete(eParser);
   }
}

// ****************************** public ******************************

void PlayerService::disconnectService()
{
   error = "";
   externalParserTerminate();
}

bool PlayerService::sendToService(QString signal)
{
   if (eParser)
     eParser->writeToStdin(signal);

   return (eParser);
}

bool PlayerService::connectService(QString parser, QString identString)
{
    error = "";

    if ( eParser )
    {
      cout << TARGET": running Player Service already running" << endl;
      eParser->writeToStdin(identString);
      return true; // should not happen
    }

    QString ppath = parsersPath;
    ParserType parserType = getParser(parser, ppath);
    
    if ( parserType == NONE )
    {
      error = "Could not find Player Service parser: " + parser;
      return false;
    }
    
    parserBuffer = "";
    
    eParser = new Q3Process( this );  
    eParser->setWorkingDirectory(QDir(ppath)); // was bug
    eParser->setCommunication( Q3Process::Stdin |
                               Q3Process::Stdout
                            );
    
    identString = identString.replace('"', "'");

    while (identString.right(1) == "\\")
      identString.truncate( identString.length() );
 
    switch (parserType)
    {
      case PERL:
			// execute perl script
			eParser->addArgument( "perl" );
			eParser->addArgument( ppath + parser );
			eParser->addArgument( "\"" + identString + "\"" );  // change
			//cout << "PERL PARSER" << endl;
			break;
		case EXEC:
			// PARSER == EXEC
			// executable binary
			eParser->addArgument( ppath + parser );
			eParser->addArgument( "\"" + identString + "\"");  // change
			//cout << "EXEC PARSER" << endl;
			break;
		case SHELL:
			// PARSER == SHELL
			// execute any executable
			//
			// requires parser files to be executable!
			// parser and identString is external
			// trust parser
			eParser->addArgument( "sh" );
			eParser->addArgument( "-c" );
			eParser->addArgument( parser + " \"" + identString + "\"" );
			//cout << "SHELL PARSER" << endl;
			break;
      case PYTHON:
			// execute perl script
			eParser->addArgument( "python" );
			eParser->addArgument( ppath + parser );
			eParser->addArgument( "\"" + identString + "\"" );  // change
			//cout << "PERL PARSER" << endl;
			break;
      default:
            break;
    }

    connect( eParser, SIGNAL(readyReadStdout()),
            this, SLOT(externalParserRead()) );
    connect( eParser, SIGNAL(processExited()),
            this, SLOT(externalParserExited()) );

    
    if ( !eParser->start() )
    {
      error = "Could not start Player Service parser: " + parser;
      return false;
    }

    return true;
}

// ****************************** private ******************************

ParserType PlayerService::checkParserType(QString &parser, QString &path)
{
    //PERL
    QFileInfo fileInfo;

    fileInfo = QFileInfo( path + parser + ".pl" );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".pl";
        return PERL;
    }

    //SHELL
    fileInfo = QFileInfo( path + parser + ".sh");
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".sh";
        return SHELL;
    }

    //EXEC
    fileInfo = QFileInfo( path + parser );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
        return EXEC;
            
    //PYTHON
    fileInfo = QFileInfo( path + parser + ".py" );
    if ( fileInfo.isFile() && fileInfo.isReadable() )
    {
        parser += ".py";
        return PYTHON;
    }

    return NONE;
}

ParserType PlayerService::getParser(QString &parser, QString &path)
{
    QString nParser = parser;
    ParserType parserType;
    
    parserType = checkParserType(nParser, path);
    
    if ( parserType != NONE )
    {
        parser = nParser;
        return parserType;
    }

    parserType = checkParserType(nParser, parsersPath_install);

    if ( parserType != NONE )
    {
        parser = nParser;
        path = parsersPath_install;
        return parserType;
    }
    
    cerr << TARGET": Player Service parser not found: " << parser.latin1() << endl;

    return NONE;
}

void PlayerService::externalParserTerminate()
{
   if ( eParser && eParser->isRunning() )
   {
      eParser->tryTerminate();
   }
}

bool PlayerService::parseScreen()
{ 
    
    QDomDocument doc;    
    
    if (!doc.setContent(parserFragment, false) )
    {
       error = "could not read Player Service output";
       
       cerr << TARGET": could not read Player Service parser output: " << endl;
       cerr << "-------------------------------------" << endl;
       cerr << parserFragment.latin1() << endl;
       cerr << "-------------------------------------" << endl;
       
       return false;
    }
    
    QDomElement docElem = doc.documentElement();
    QDomNode fields = docElem.firstChild();

    QDomNode fName, fType, fValue;
    QString name, type, value;
        
    while( !fields.isNull() )
    {
        QDomElement e = fields.toElement();
        
        if ( e.isNull() || e.tagName() != "field" )
        {
          error = "could not read Player Service output";
          cerr << TARGET": invalid Player Service parser output" << endl;
          return false;
        }
       
        fName  = e.elementsByTagName("name" ).item(0);
        fType  = e.elementsByTagName("type"  ).item(0);
        fValue = e.elementsByTagName("value").item(0);
        
        if ( !fName.isNull() && !fType.isNull() )  // required nodes
        {
              name = fName.toElement().text();
              type  = fType.toElement().text();
              
              if ( fValue.isNull() )  // value
                value = "";
              else
                value = fValue.toElement().text();
                
             value = QString::fromUtf8(value);
             emit playerServiceParameter(name, type, value);
        }
          else
        {
           error = "could not read Player Service output";
           cerr << TARGET": invalid Player Service parser output" << endl;
        }
                  
        fields = fields.nextSibling();
    }

    return true;
}        

void PlayerService::externalParserRead()
{
   parserFragment = QString::null;
   do 
   {
     if (parserFragment != QString::null)
     { 
        parserBuffer += parserFragment;
        //cout << temp << endl;
     }
     parserFragment = eParser->readLineStdout();
   }
     while (parserFragment != QString::null);

   int pos = parserBuffer.find("</screen>", 0, false);

   if (pos > -1)
   {
     parserFragment = parserBuffer.left(pos + 9);
     parserBuffer = parserBuffer.mid(pos + 9);
     if ( !parseScreen() )
       emit playerServiceDataError(error);
   }
}


void PlayerService::externalParserExited()
{
   if (parserBuffer != "")
   {
     parserFragment = parserBuffer;
     parseScreen();
   }
   if (eParser)
     delete eParser;
   
   eParser = NULL;
}

