/* ============================================================
 * File  : streamtuned.h
 * Author: Eric Giesselbach <ericgies@kabelfoon.nl>
 * Date  : 2004-02-12
 * Description : streamtuned gui
 *
 * Copyright 2003 by Eric Giesselbach

 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published bythe Free Software Foundation;
 * either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ============================================================ */

#ifndef MYTHSTREAM_H
#define MYTHSTREAM_H

#include <qwidget.h>
#include <qdatetime.h>
#include <q3process.h>
//Added by qt3to4:
#include <QPaintEvent>
#include <QPixmap>
#include <QKeyEvent>
#include <QEvent>

#include <mythtv/uitypes.h>
#include <mythtv/xmlparse.h>
#include <mythtv/oldsettings.h>
#include <mythtv/mythwidgets.h>
#include <mythtv/mythdialogs.h>

#include <qrect.h>
#include <qtimer.h>
#include <qbitmap.h>

#include "../libs/streambrowser.h"
#include "../libs/streamconfig.h"
#include "../libs/storageconfig.h"
#include "../libs/streamstorage.h"

#include "../libs/speechproc.h"


class FFTBox : public QWidget
{
   Q_OBJECT

   public:
     FFTBox(QWidget *parent, int barcnt, unsigned int x, unsigned int y, unsigned int w, unsigned int h);
     ~FFTBox();

     void resetDisplay();
     void paintEvent( QPaintEvent * );

   public slots:
     void setPercentage(Spectrum* percent);

   private slots:
     void buildFFTBackground();

   private:
     Spectrum *percentage;
     int barCount;
     int barWidth;
     int gaugeWidth;
     float div;
     bool go;

     QPixmap bckPict, fftPict, drawPict;
     QBitmap fftMask;

     QTimer *grabTimer;

};


class PlayerState : public QObject
{
   Q_OBJECT

   public:

     // displayStatus indicates state of display. 
     // Allowed values: ds_browse, ds_audio, ds_video, ds_browse + ds_audio, ds_browse + ds_video
     const static uint ds_browse = 1; // video container, video panel, audio panel not visible
     const static uint ds_audio  = 2; // audio playing
     const static uint ds_video  = 4; // video playing (videoStatus != novideo)
     
     // videoStatus indicates state of video container
     enum VideoStatus
     {
        novideo = 0, 
        preview,
        fullscreen,
        zoomed
     };
     
     PlayerState();

     void streamStatusChanged(QString status);
     void browserActivityDetected();
     void videoStatusChanged(VideoStatus videoStatus);

     bool pollMe(); // returns true on displayState change

     QString     getStreamStatus()  { return streamStatus;  }
     uint        getDisplayStatus() { return displayStatus; }
     VideoStatus getVideoStatus()   { return videoStatus;  }

   private:

     QString streamStatus;
     VideoStatus videoStatus;
     uint displayStatus;

     int displayResetDelay;
};

class MythStream : public MythDialog
{
    Q_OBJECT

  enum Section
  {
     top = 0,
     mid,
     bot,
     all
  };

  enum FieldOwner
  {
     o_harvester = 0,
     o_player,
     o_playerservice,
     o_streambrowser,
     o_all
  };

  public:

    MythStream(MythMainWindow *parent,
               const char *name = 0);
    ~MythStream();
  
  protected:
    bool eventFilter( QObject *obj, QEvent *ev);

  private slots:
    void slotItemTreeSwitchedTo(int itemTreeLoaded);
    void slotHarvesterBusy(bool busy, QString message);
    void slotVideoStatusChanged(int videoStatus);
    void slotUserMessage(QString&, QString&);
    void slotGuardInfoField();
    void slotValuesUpdated(int updatedType);
    void slotStorageEvent(int ident, int eventType, bool error);
    void slotWebStorageMaybeReady();
    void slotEditString(QString &line, QString label, bool& cancelled);
    void slotPopResponse(void);
    void slotPopStreamEditBox(QString &errorMessage);
    void slotPlayerServiceParameter(QString, QString, QString);
    void slotPlayerServiceDataError(QString);
    void slotFetchReady(bool, QString&); //requester used for image download
    
    void playerButton1Pushed();
    void playerButton2Pushed();
    void playerButton3Pushed();
    void playerButtonsReset();

  private:
    void LoadWindow(QDomElement &element);
    void updateBackground();
    void paintEvent(QPaintEvent *e);

    void getStreamStatusValues();
    void getStreamPlayValues();
    void checkShowPlayer();
    void getFolderList();
    void getItemList();
    void displayDetails();


    QString getCurrentItemName(QString type);

    //void parseFont(QDomElement& e);
    void parseContainer(QDomElement& e, QString& name, QRect& area);

    void invalidateSection(Section sect);
    void updateInvalidated();

    bool topSectInvalidated;
    bool midSectInvalidated;
    bool botSectInvalidated;
    
    void updateTopView();
    void updateMidView();
    void updateBotView();

    void keyPressEvent( QKeyEvent *k );

    void loadUIImageType(UIImageType *icon, QChar prefix);
    void loadField(const QString& container, const QString& field, QString& value);
    void loadListFields(const QString& container, const QString& fieldbasename, QStringList& value, int cursor);
    void loadBar(const QString& container, const QString& field, int value);
    void loadIconSource(QString containerName, QString name, QPixmap &source);
    UITextType * findTtype(const QString field);
    bool processAction( QString action );

    void selectStorageByIndex(ReposStorage *storageBin, StreamStorage *streamStorage, int index);

    void toggleSpeech();
    void stopSpeech();

    void playerButtonPush(int butnum);

    QString popParameterBox(QString myString, QString label, bool &cancelled);
    
    SpeechProc *speechProc;
    
    XMLParse     *m_Theme;
    QRect        m_TopRect;
    QRect        m_MidRect;
    QRect        m_BotRect;
    QRect        m_VideoRect;
    QRect        m_VideoMaxRect;
    QRect        m_ViewerRect;
    QRect        m_ViewerMaxRect;
    unsigned int m_ItemDisplaySize;
    unsigned int m_FolderDisplaySize;
    bool harvesting;
    
    FieldOwner messageOwner;
    FieldOwner custominfoOwner;
    FieldOwner logoOwner;
    
    QRect m_fftRect;
    FFTBox *fft;

    StreamBrowser *streamBrowser;

    void reportEvent(QString msg); // message clears in 4 sec
    void setMessage(QString msg);  // "permanent" message

    // state
    StreamBrowser::ItemTreeLoaded itemTreeLoaded;
    PlayerState playerState;
    bool playerServiceActive; // audio_panel and logo updates to player service
 
    QTimer *infoTimer;   // clears message and info field
    bool infoTimerRerun; // message came in during timer pending, rerun timer on timeout

    StreamStorage *streamStorage;
    ReposStorage *storageBin;

    int displayedVolume;
    int commandPending; // state <>0 means input expected

    // popup boxes
    MythPopupBox *popBox;
    MythLineEdit *editBox;
    QObject *buttonPushed;
    
    // stream item icons
    QPixmap streamPict, downloadPict, urlPict, unknownPict, scheduledPict, emptyPict,
            recordingPict, recordedPict, generalPict, headerPict, markedPict,
            leftPict, rightPict, infoPict, htmlPict;

    UIImageType *logo;
    QString currentLogoUrl; 

    UITextButtonType *playerButton1, *playerButton2, *playerButton3;

    Requester *requester;
    
    QPixmap bgPixmap;
};

// ----------------------------------------------------------------------

void selectStorages(ReposStorage *storageBin, StreamStorage *streamStorage, int ident = 0);



#endif /* STREAMTUNED_H */
